
@file:Suppress("unused", "UNUSED_PARAMETER")

package ru.dbotthepony.mc.otm

import net.minecraft.server.MinecraftServer
import net.minecraft.world.level.Level
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.event.TickEvent.LevelTickEvent
import net.minecraftforge.event.server.ServerAboutToStartEvent
import net.minecraftforge.event.server.ServerStoppedEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import net.minecraftforge.eventbus.api.EventPriority
import net.minecraftforge.eventbus.api.SubscribeEvent
import org.apache.logging.log4j.LogManager
import java.util.*
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayList

private class TickList {
	private val conditional = LinkedList<IConditionalTickable>()
	private val once = LinkedList<ITickable>()

	private val conditionalValveTime = LinkedList<IConditionalTickable>()
	private val onceValveTime = LinkedList<ITickable>()

	private var inTicker = false

	fun add(ticker: IConditionalTickable) {
		if (inTicker) {
			conditionalValveTime.add(ticker)
		} else {
			conditional.addFirst(ticker)
		}
	}

	fun add(ticker: ITickable) {
		if (inTicker) {
			onceValveTime.add(ticker)
		} else {
			once.addFirst(ticker)
		}
	}

	fun tick() {
		if (inTicker) {
			throw ConcurrentModificationException("Already ticking")
		}

		inTicker = true
		val iterator = conditional.iterator()

		for (ticker in iterator) {
			if (!ticker.canTick) {
				iterator.remove()
			} else {
				ticker.tick()
			}
		}

		for (ticker in once) {
			ticker.tick()
		}

		once.clear()

		for (ticker in conditionalValveTime) {
			conditional.addFirst(ticker)
		}

		for (ticker in onceValveTime) {
			once.addFirst(ticker)
		}

		conditionalValveTime.clear()
		onceValveTime.clear()

		inTicker = false
	}

	fun clear() {
		conditional.clear()
		once.clear()
	}
}

private val preServerTick = TickList()
private val postServerTick = TickList()
private val preWorldTick = WeakHashMap<Level, TickList>()
private val postWorldTick = WeakHashMap<Level, TickList>()

private val preServerTickTimers = TimerQueue()
private val postServerTickTimers = TimerQueue()

class TimerQueue {
	private var ticks = 0
	private var head: Timer? = null

	inner class Timer(val timerTicks: Int, val runnable: Runnable) {
		val ringAt = ticks + timerTicks
		var finished = false
			private set

		var next: Timer? = null
		var prev: Timer? = null

		init {
			if (head == null) {
				head = this
			} else {
				var next = head

				while (next != null) {
					if (next.ringAt >= this.ringAt) {
						next.prev?.next = this
						this.prev = next.prev
						next.prev = this
						this.next = next
						break
					} else if (next.next == null) {
						next.next = this
						this.prev = next
						break
					} else {
						next = next.next
					}
				}
			}
		}

		fun execute() {
			check(!finished) { "Already finished" }
			runnable.run()
			finished = true
		}
	}

	fun tick() {
		ticks++

		var head = head

		while (head != null) {
			if (head.ringAt <= ticks) {
				head.execute()
				head = head.next
				head?.prev = null
				this.head = head
			} else {
				break
			}
		}
	}

	fun clear() {
		ticks = 0
		head = null
	}
}

fun addPreTickTimer(inTicks: Int, callback: Runnable): TimerQueue.Timer? {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add timer $callback in ticks $inTicks while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return preServerTickTimers.Timer(inTicks, callback)
}

fun addPostTickTimer(inTicks: Int, callback: Runnable): TimerQueue.Timer? {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $callback in ticks $inTicks while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return postServerTickTimers.Timer(inTicks, callback)
}

private var _server: MinecraftServer? = null
private var _serverThread: Thread? = null
private var _clientThread: Thread? = null

val isClient: Boolean get() = _clientThread !== null

fun recordClientThread() {
	if (_clientThread != null) {
		throw IllegalStateException("Already have client channel")
	}

	_clientThread = Thread.currentThread()
}

fun runIfClient(lambda: () -> Unit) {
	if (isClient) {
		lambda.invoke()
	}
}

fun <V> runIfClient(value: V, lambda: () -> V): V {
	if (isClient) {
		return lambda.invoke()
	}

	return value
}

fun <V> runOnClient(value: V, lambda: () -> V): V {
	if (isClientThread()) {
		return lambda.invoke()
	}

	return value
}

fun isServerThread(): Boolean {
	return Thread.currentThread() === _serverThread
}

fun isClientThread(): Boolean {
	return Thread.currentThread() === _clientThread
}

val MINECRAFT_SERVER: MinecraftServer
	get() {
		return _server ?: throw NullPointerException("Not running a server!")
	}

val NULLABLE_MINECRAFT_SERVER: MinecraftServer?
	get() {
		return _server
	}

// Flag whenever is server alive
// To avoid baby deadlocks caused by minecraft engine design issues
var SERVER_IS_DYING = true
	private set

private val LOGGER = LogManager.getLogger()

fun interface ITickable {
	fun tick()
}

interface IConditionalTickable : ITickable {
	/**
	 * Once this returns false, it should stay false.
	 *
	 * If it suddenly turns true after being false, result is undefined.
	 */
	val canTick: Boolean
}

fun onServerTick(event: ServerTickEvent) {
	if (event.phase === TickEvent.Phase.START) {
		preServerTickTimers.tick()
		preServerTick.tick()
	} else {
		postServerTickTimers.tick()
		postServerTick.tick()
	}
}

fun addPreServerTicker(ticker: IConditionalTickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preServerTick.add(ticker)
}

fun addPostServerTicker(ticker: IConditionalTickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postServerTick.add(ticker)
}

fun addPreServerTickerOnce(ticker: ITickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preServerTick.add(ticker)
}

fun addPostServerTickerOnce(ticker: ITickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postServerTick.add(ticker)
}

fun onWorldTick(event: LevelTickEvent) {
	if (event.phase === TickEvent.Phase.START) {
		preWorldTick[event.level]?.tick()
	} else {
		postWorldTick[event.level]?.tick()
	}
}

fun addPreWorldTicker(level: Level, ticker: IConditionalTickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preWorldTick.computeIfAbsent(level) { TickList() }.add(ticker)
}

fun addPostWorldTicker(level: Level, ticker: IConditionalTickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postWorldTick.computeIfAbsent(level) { TickList() }.add(ticker)
}

fun addPreWorldTickerOnce(level: Level, ticker: ITickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preWorldTick.computeIfAbsent(level) { TickList() }.add(ticker)
}

fun addPostWorldTickerOnce(level: Level, ticker: ITickable) {
	if (SERVER_IS_DYING) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postWorldTick.computeIfAbsent(level) { TickList() }.add(ticker)
}

private fun clear() {
	preServerTickTimers.clear()
	postServerTickTimers.clear()
	preServerTick.clear()
	postServerTick.clear()
	preWorldTick.clear()
	postWorldTick.clear()
}

fun onServerStarting(event: ServerAboutToStartEvent) {
	clear()
	SERVER_IS_DYING = false
	_server = event.server
	_serverThread = Thread.currentThread()
}

fun onServerStopping(event: ServerStoppingEvent) {
	clear()
	SERVER_IS_DYING = true
}

fun onServerStopped(event: ServerStoppedEvent) {
	if (!SERVER_IS_DYING) {
		LOGGER.fatal("ServerStoppingEvent did not fire. If server has crashed this is normal. However, if server finished it's work 'gracefully' this is a bug.")

		clear()
		SERVER_IS_DYING = true
	}

	_server = null
	_serverThread = null
}
