package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentContents
import net.minecraft.network.chat.MutableComponent
import net.minecraft.network.chat.contents.TranslatableContents
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.core.getKeyNullable
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.util.*
import kotlin.collections.HashSet
import kotlin.collections.RandomAccess

fun interface AndroidResearchFactory<R : AndroidResearch> {
	fun factory(type: AndroidResearchType<*>, capability: MatteryPlayerCapability): R
}

private fun findPrerequisites(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
	top: Boolean = true
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		if (!top) {
			add.add(value)
		}

		findPrerequisites(value.definedPrerequisites, add, false)
	}

	return add
}

private fun findAllPrerequisites(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		add.add(value)
		findAllPrerequisites(value.flatPrerequisites, add)
	}

	return add
}

private fun findAllChildren(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		add.add(value)
		findAllChildren(value.flatUnlocks, add)
	}

	return add
}

class ListSet<T>(private val list: ImmutableList<T>) : List<T>, Set<T>, RandomAccess {
	constructor(list: Collection<T>) : this(ImmutableList.copyOf(list))

	private val set: ImmutableSet<T> by lazy { ImmutableSet.copyOf(list) }

	override val size: Int
		get() = list.size

	override fun contains(element: T): Boolean {
		return set.contains(element)
	}

	override fun containsAll(elements: Collection<T>): Boolean {
		return set.containsAll(elements)
	}

	override fun get(index: Int): T {
		return list[index]
	}

	override fun indexOf(element: T): Int {
		return list.indexOf(element)
	}

	override fun isEmpty(): Boolean {
		return list.isEmpty()
	}

	override fun iterator(): Iterator<T> {
		return list.iterator()
	}

	override fun lastIndexOf(element: T): Int {
		return list.lastIndexOf(element)
	}

	override fun listIterator(): ListIterator<T> {
		return list.listIterator()
	}

	override fun listIterator(index: Int): ListIterator<T> {
		return list.listIterator(index)
	}

	override fun spliterator(): Spliterator<T> {
		return list.spliterator()
	}

	override fun subList(fromIndex: Int, toIndex: Int): List<T> {
		return list.subList(fromIndex, toIndex)
	}
}

open class AndroidResearchType<R : AndroidResearch>(
	protected val factory: AndroidResearchFactory<R>
) {
	val researchTreeDepth: Int by lazy {
		if (flatPrerequisites.isEmpty()) {
			return@lazy 0
		}

		var depth = 1

		for (value in flatPrerequisites) {
			depth = depth.coerceAtLeast(value.researchTreeDepth + 1)
		}

		return@lazy depth
	}

	/**
	 * Prerequisites as-is.
	 *
	 * Please avoid having more than one prerequisite, as this case doesn't have proper research tree
	 * rendering code (yet).
	 */
	open val definedPrerequisites: List<AndroidResearchType<*>> get() = emptyList()

	/**
	 * Blocked by list as-is
	 */
	open val definedBlockedBy: List<AndroidResearchType<*>> get() = emptyList()

	/**
	 * Flat list of research preceding this research.
	 *
	 * This list won't contain research following the same research path twice (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B depends on A
	 *
	 * C specify both B and A as it's prerequisites, [flatPrerequisites] will contain only B, when [definedPrerequisites] will contain
	 * both B and A
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatPrerequisites: List<AndroidResearchType<*>> by lazy {
		val parentPrerequisites = findPrerequisites(definedPrerequisites)
		val builder = ImmutableList.builder<AndroidResearchType<*>>()

		for (value in definedPrerequisites) {
			if (value !in parentPrerequisites) {
				builder.add(value)
			}
		}

		ListSet(builder.build())
	}

	/**
	 * All prerequisite research, including indirect ones
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allPrerequisites: List<AndroidResearchType<*>> by lazy {
		ListSet(findAllPrerequisites(flatPrerequisites))
	}

	/**
	 * Flat list of research blocked by this research.
	 *
	 * This list won't contain children of any blocked research (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B is blocked by A
	 * * C is blocked by A
	 *
	 * Both C and B specify A as it's blockers, [flatBlocking] will contain only B, because C is blocked by A through B.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatBlocking: List<AndroidResearchType<*>> by lazy {
		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (this in research.definedBlockedBy) {
				var hit = false

				for (parent in research.allPrerequisites) {
					if (this in parent.definedBlockedBy) {
						hit = true
						break
					}
				}

				if (!hit) {
					list.add(research)
				}
			}
		}

		ListSet(list.build())
	}

	/**
	 * Flat list of research blocking this research.
	 *
	 * This list won't contain any research blocked by this' research parents.
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B is blocked by A
	 * * C is blocked by A
	 *
	 * Both B and C are blocked by A, but C requires B, so C's [flatBlockedBy] will not contain A, but B's will.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatBlockedBy: ListSet<AndroidResearchType<*>> by lazy {
		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (blocker in definedBlockedBy) {
			var hit = false

			for (research in allPrerequisites) {
				if (blocker in research.flatBlockedBy) {
					hit = true
					break
				}
			}

			if (!hit) {
				list.add(blocker)
			}
		}

		ListSet(list.build())
	}

	/**
	 * All research blocking this research, even indirect ones (e.g. blocked by through one of parent research).
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allBlockedBy: ListSet<AndroidResearchType<*>> by lazy {
		val list = HashSet<AndroidResearchType<*>>()

		list.addAll(flatBlockedBy)

		for (parent in allPrerequisites) {
			list.addAll(parent.flatBlockedBy)
		}

		ListSet(list)
	}

	/**
	 * All research directly unlocked by this research.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatUnlocks: List<AndroidResearchType<*>> by lazy {
		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (this in research.flatPrerequisites) {
				list.add(research)
			}
		}

		ListSet(list.build())
	}

	/**
	 * All research unlocked by this research, including indirect ones
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allUnlocks: List<AndroidResearchType<*>> by lazy {
		ListSet(findAllChildren(flatUnlocks))
	}

	/**
	 * All research blocked by this research, including indirect ones.
	 *
	 * Returns list which also doubles as set (for contains method)
	 */
	val allBlocking: List<AndroidResearchType<*>> by lazy {
		val set = HashSet<AndroidResearchType<*>>()

		for (research in flatBlocking) {
			set.add(research)
			set.addAll(research.allUnlocks)
		}

		for (research in allPrerequisites) {
			for (research2 in research.flatBlocking) {
				set.add(research2)
				set.addAll(research2.allUnlocks)
			}
		}

		ListSet(set)
	}

	fun factory(capability: MatteryPlayerCapability) = factory.factory(this, capability)

	val registryName by lazy {
		MRegistry.ANDROID_RESEARCH.getKeyNullable(this)
	}

	val displayId by lazy {
		val registryName = registryName ?: throw NullPointerException("No registry name present")
		return@lazy "android_research.${registryName.namespace}.${registryName.path}".intern()
	}

	val descriptionId by lazy {
		return@lazy "$displayId.description".intern()
	}

	open val displayContents: ComponentContents by lazy {
		TranslatableContents(displayId)
	}

	open val displayName: Component by lazy {
		MutableComponent.create(displayContents)
	}

	open val displayDescription: List<Component> by lazy {
		listOf(TranslatableComponent(descriptionId))
	}
}
