package ru.dbotthepony.mc.otm.matter

import net.minecraft.tags.ItemTags
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.item.Items
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeType
import net.minecraftforge.event.server.ServerStartedEvent
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.tags.ITag
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.registry.MItems
import ru.dbotthepony.mc.otm.registry.MRecipes
import kotlin.math.pow
import kotlin.math.sqrt

/**
 * baselineComplexityReplicateTicks * complexity = replicate ticks
 * baselineComplexityDecomposeTicks * complexity = decompose ticks
 */
const val baselineComplexityReplicateTicks = 1000.0
const val baselineComplexityDecomposeTicks = 400.0
const val baselineComplexityScanTicks = 4000.0

@JvmField val BASELINE = ImpreciseFraction(0.001)
@JvmField val DIRT = BASELINE
@JvmField val STONY = BASELINE * 0.92

@JvmField val IRON = BASELINE * 60
@JvmField val COPPER = BASELINE * 36
@JvmField val GOLD = IRON * 2.5
@JvmField val LAPIS = IRON * 1.15
@JvmField val REDSTONE = IRON * 1.4
@JvmField val DIAMOND = GOLD * 4
@JvmField val EMERALD = DIAMOND
@JvmField val QUARTZ = IRON * 0.97
@JvmField val COAL = DIAMOND * 0.08
@JvmField val TRITANIUM = IRON * 1.95

@JvmField val WATER = BASELINE * 0.3

@JvmField val AMETHYST = EMERALD * 0.44

@JvmField val NETHER_STAR = ImpreciseFraction(40)

@JvmField val MUSIC_DISC = BASELINE * 200

@JvmField val SOUL_SAND = BASELINE * 6.66666666
@JvmField val GLOWSTONE_DUST = GOLD * 0.44 + REDSTONE * 0.7

@JvmField val SEEDS = BASELINE * 1.8
@JvmField val PLANTS = BASELINE * 4.6
@JvmField val GRASS = PLANTS * 0.65

@JvmField val SIMPLE_ORGANICS = BASELINE * 9
@JvmField val STRINGS = SIMPLE_ORGANICS * 0.8
@JvmField val FEATHERS = SIMPLE_ORGANICS * 1.2
@JvmField val BONES = SIMPLE_ORGANICS * 1.8

@JvmField val ADVANCED_ORGANICS = SIMPLE_ORGANICS * 4.2

@JvmField val PRISMARINE = IRON * 0.44

@JvmField val LOGS = BASELINE * 4
@JvmField val STRIPPED_LOGS = LOGS * 0.91
@JvmField val PLANKS = LOGS * 6

// трава?
@JvmField val CORALS = ADVANCED_ORGANICS * 0.83
@JvmField val DEAD_CORALS = CORALS * 0.91

private val seenItems = ArrayDeque<Item>()
private val seenRecipes = ArrayDeque<Recipe<*>>()
private val LOGGER = LogManager.getLogger()
private val mappedOutputs = HashMap<Item, ArrayList<Recipe<*>>>()
private val mappedInputs = HashMap<Item, ArrayList<Recipe<*>>>()

private fun defaultComplexity(value: ImpreciseFraction): Double {
	val double = value.toDouble()

	if (double < 1.0) {
		return sqrt(double)
	}

	return double.pow(1.4)
}

private fun make(item: Item, value: ImpreciseFraction, complexity: Double = defaultComplexity(value)) {
	rootEntriesAccess[item] = MatterTuple(value, complexity)
}

private fun derive(from: Item, to: Item, multValue: ImpreciseFraction = ImpreciseFraction.ONE, multComplexity: Double = 1.0) {
	val getValue = getMatterValueNullable(from)

	if (getValue != null) {
		derivedEntriesAccess[to] = MatterTuple(getValue.value * multValue, getValue.complexity * multComplexity)
	}
}

private fun derive(from: Item, to: Item, multValue: Double, multComplexity: Double = 1.0) = derive(from, to, ImpreciseFraction(multValue), multComplexity)

private fun deriveIfMissing(from: Item, to: Item, multValue: ImpreciseFraction = ImpreciseFraction.ONE, multComplexity: Double = 1.0) {
	if (getMatterValueNullable(to) != null)
		return

	val getValue = getMatterValueNullable(from)

	if (getValue != null) {
		derivedEntriesAccess[to] = MatterTuple(getValue.value * multValue, getValue.complexity * multComplexity)
	}
}

private fun deriveIfMissing(from: Item, to: Item, multValue: Double, multComplexity: Double = 1.0) = deriveIfMissing(from, to, ImpreciseFraction(multValue), multComplexity)

private fun make(tag: ITag<Item>, value: ImpreciseFraction, complexity: Double = defaultComplexity(value)) {
	for (item in tag.stream().iterator()) {
		make(item, value, complexity)
	}
}

private fun make(tag: TagKey<Item>, value: ImpreciseFraction, complexity: Double = defaultComplexity(value)) {
	val resolved = ForgeRegistries.ITEMS.tags()?.getTag(tag)

	if (resolved != null) {
		make(resolved, value, complexity)
	}
}

private fun logResolutionError(text: String) {
	LOGGER.debug("... Item chain: " + seenItems.joinToString(" -> "))
	// LOGGER.error("Recipe chain: " + seenRecipes.joinToString(" -> "))
	LOGGER.debug(text)
}

private enum class ResultType {
	RESOLVED,
	RECURSION,
	UNKNOWN_VALUE
}

private data class Result(val type: ResultType = ResultType.UNKNOWN_VALUE, val tuple: MatterTuple? = null)

fun serverStartData(event: ServerStartedEvent) {
	rootEntriesAccess.clear()
	derivedEntriesAccess.clear()
	mappedOutputs.clear()
	mappedInputs.clear()

	building = true
	registerRoots()

	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(RecipeType.SMELTING))
	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(RecipeType.BLASTING))
	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(RecipeType.SMOKING))
	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(RecipeType.STONECUTTING))
	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(MRecipes.PLATE_PRESS))
	floodRecipeCategory(event.server.recipeManager.getAllRecipesFor(RecipeType.CRAFTING))

	for (item in mappedInputs.keys) {
		determineMatterValue(item)
	}

	for (item in mappedOutputs.keys) {
		determineMatterValue(item)
	}

	registerDerived()
	building = false
}

private fun floodRecipeCategory(list: Collection<Recipe<*>>) {
	for (recipe in list) {
		val result = recipe.resultItem

		if (!result.isEmpty) {
			mappedOutputs.computeIfAbsent(result.item) {ArrayList()}.add(recipe)

			for (inputI in recipe.ingredients) {
				for (stack in inputI.items) {
					if (!stack.isEmpty) {
						mappedInputs.computeIfAbsent(stack.item) {ArrayList()}.add(recipe)
					}
				}
			}
		}
	}
}

private fun determineMatterValue(item: Item): Result {
	val rootEntry = getMatterValueNullable(item)

	if (rootEntry != null)
		return Result(ResultType.RESOLVED, rootEntry)

	val recipes = mappedOutputs[item] ?: return Result(ResultType.UNKNOWN_VALUE)

	if (seenItems.contains(item)) {
		// logResolutionError("Already seen this item: $item")
		return Result(ResultType.RECURSION)
	}

	seenItems.addLast(item)
	var minimal: MatterTuple? = null
	var anyRecursion = false

	for (recipe in recipes) {
		if (seenRecipes.contains(recipe)) {
			logResolutionError("Already seen this recipe: $recipe")
			continue
		}

		seenRecipes.addLast(recipe)
		var sum = MatterTuple.ZERO
		var recursiveRecipe = false
		var once = false

		for (ingredient in recipe.ingredients) {
			var minIngredient: MatterTuple? = null
			var recursive = false

			for (stack in ingredient.items) {
				val getStackTuple = determineMatterValue(stack.item)

				if (getStackTuple.type == ResultType.UNKNOWN_VALUE) {
					logResolutionError("Failed to resolve $item value because ${stack.item} is unknown")
					seenItems.removeLast()
					seenRecipes.removeLast()
					return getStackTuple
				} else if (getStackTuple.type == ResultType.RESOLVED) {
					if (minIngredient == null || minIngredient > getStackTuple.tuple!!)
						minIngredient = getStackTuple.tuple
				} else {
					recursive = true
				}
			}

			//if (minIngredient == null) {
			//	if (recursive) {
			//		recursiveRecipe = true
			//		break
			//	}

			//	logResolutionError("Failed to resolve $item value because $ingredient is empty (${ingredient.items.joinToString(", ")})???")
			//	seenItems.removeLast()
			//	seenRecipes.removeLast()
			//	return Result(ResultType.UNKNOWN_VALUE)
			//}

			if (recursive) {
				recursiveRecipe = true
			}

			if (minIngredient != null) {
				sum += minIngredient
				once = true
			}
		}

		seenRecipes.removeLast()

		if (recursiveRecipe) {
			anyRecursion = true
			// logResolutionError("$recipe is anything but recursion?")
			// continue
		}

		if (!once) {
			if (recursiveRecipe) {
				//logResolutionError("Escaping resolution of $item early because $recipe is recursive")
				//seenItems.removeLast()
				//return Result(ResultType.RECURSION)
				continue
			}

			logResolutionError("Failed to resolve $item because $recipe is unresolvable")
			seenItems.removeLast()
			return Result(ResultType.UNKNOWN_VALUE)
		}

		sum = MatterTuple(sum.value / recipe.resultItem.count, sum.complexity / recipe.resultItem.count)

		if (minimal == null || minimal > sum) {
			minimal = sum
		}
	}

	seenItems.removeLast()

	if (minimal != null) {
		derivedEntriesAccess[item] = minimal
		return Result(ResultType.RESOLVED, minimal)
	}

	if (anyRecursion) {
		logResolutionError("Failed to resolve $item because hit recursive recipes")
		return Result(ResultType.RECURSION)
	}

	logResolutionError("Failed to resolve $item because no appropriate recipes were found")
	return Result()
}

private fun registerRoots() {
	make(Items.STONE, STONY)
	make(Items.COBBLESTONE, STONY)
	make(Items.GRAVEL, STONY)
	make(Items.TUFF, STONY * 1.17)
	make(Items.CALCITE, STONY * 1.17)
	make(Items.SAND, STONY)
	make(Items.RED_SAND, STONY)
	make(Items.FLINT, STONY)
	make(Items.NETHERRACK, STONY * 0.7)
	make(Items.SOUL_SAND, SOUL_SAND)
	make(Items.SOUL_SOIL, SOUL_SAND)
	make(Items.DEEPSLATE, STONY * 1.15)
	make(Items.COBBLED_DEEPSLATE, STONY * 1.15)
	make(Items.BLACKSTONE, STONY * 1.21)
	make(Items.OBSIDIAN, STONY * 3.5)

	make(Items.INFESTED_COBBLESTONE, STONY + ADVANCED_ORGANICS * 3.6)
	make(Items.MOSSY_COBBLESTONE, STONY + GRASS * 0.6)

	make(Items.SNOWBALL, WATER / 4)

	make(ItemTags.PLANKS, PLANKS)

	make(ItemTags.LOGS, LOGS)

	make(ItemTags.SAPLINGS, PLANTS * 2.6)

	make(Items.STRIPPED_OAK_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_SPRUCE_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_BIRCH_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_JUNGLE_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_ACACIA_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_DARK_OAK_LOG, STRIPPED_LOGS)
	make(Items.STRIPPED_CRIMSON_STEM, STRIPPED_LOGS)
	make(Items.STRIPPED_WARPED_STEM, STRIPPED_LOGS)
	make(Items.STRIPPED_OAK_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_SPRUCE_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_BIRCH_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_JUNGLE_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_ACACIA_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_DARK_OAK_WOOD, STRIPPED_LOGS)
	make(Items.STRIPPED_CRIMSON_HYPHAE, STRIPPED_LOGS)
	make(Items.STRIPPED_WARPED_HYPHAE, STRIPPED_LOGS)

	make(Items.DIRT, DIRT)
	make(Items.DIRT_PATH, DIRT)
	make(Items.GRASS_BLOCK, DIRT * 1.1)
	make(Items.PODZOL, DIRT * 1.15)

	make(Items.COAL, COAL)
	make(Items.QUARTZ, QUARTZ)
	make(Items.AMETHYST_SHARD, AMETHYST)
	make(Items.DIAMOND, DIAMOND)
	make(Items.IRON_INGOT, IRON)
	make(MItems.TRITANIUM_INGOT, TRITANIUM)
	make(Items.COPPER_INGOT, COPPER)
	make(Items.LAPIS_LAZULI, LAPIS)
	make(Items.REDSTONE, REDSTONE)
	make(Items.GOLD_INGOT, GOLD)
	make(Items.EMERALD, EMERALD)
	make(Items.NETHER_STAR, NETHER_STAR)
	make(Items.GLOWSTONE_DUST, GLOWSTONE_DUST)

	make(Items.GUNPOWDER, BASELINE * 2.4)
	make(Items.CLAY_BALL, BASELINE * 1.9)
	make(Items.PRISMARINE_CRYSTALS, PRISMARINE * 1.3)
	make(Items.PRISMARINE_SHARD, PRISMARINE)

	make(Items.WHEAT, PLANTS)
	make(Items.BEETROOT, PLANTS * 1.1)
	make(Items.NETHER_WART, PLANTS * 2.4)
	make(Items.WHEAT_SEEDS, SEEDS)
	make(Items.MELON_SEEDS, SEEDS * 1.1)
	make(Items.PUMPKIN_SEEDS, SEEDS * 0.9)
	make(Items.BEETROOT_SEEDS, SEEDS * 1.26)
	make(Items.SUGAR, PLANTS * 0.75)
	make(Items.SUGAR_CANE, PLANTS * 1.55)
	make(Items.BAMBOO, PLANTS * 1.45)
	make(Items.CACTUS, PLANTS * 1.8)
	make(Items.VINE, PLANTS * 0.9)
	make(Items.GLOW_LICHEN, PLANTS * 1.35)
	make(Items.LILY_PAD, PLANTS)
	make(Items.SPORE_BLOSSOM, PLANTS * 1.4)
	make(Items.CHORUS_FRUIT, PLANTS * 1.55)
	make(Items.CHORUS_PLANT, PLANTS * 1.55)
	make(Items.CHORUS_FLOWER, PLANTS * 1.55)

	make(Items.SUNFLOWER, PLANTS * 1.5)
	make(Items.LILAC, PLANTS * 1.5)
	make(Items.ROSE_BUSH, PLANTS * 1.5)
	make(Items.PEONY, PLANTS * 1.5)
	make(Items.DANDELION, PLANTS * 1.24)
	make(Items.POPPY, PLANTS * 1.24)
	make(Items.BLUE_ORCHID, PLANTS * 1.24)
	make(Items.ALLIUM, PLANTS * 1.24)
	make(Items.AZURE_BLUET, PLANTS * 1.24)
	make(Items.RED_TULIP, PLANTS * 1.24)
	make(Items.ORANGE_TULIP, PLANTS * 1.24)
	make(Items.WHITE_TULIP, PLANTS * 1.24)
	make(Items.PINK_TULIP, PLANTS * 1.24)
	make(Items.OXEYE_DAISY, PLANTS * 1.24)
	make(Items.CORNFLOWER, PLANTS * 1.24)
	make(Items.LILY_OF_THE_VALLEY, PLANTS * 1.24)

	make(Items.WITHER_ROSE, PLANTS * 2.35)

	make(Items.BROWN_MUSHROOM, PLANTS)
	make(Items.RED_MUSHROOM, PLANTS)
	make(Items.COCOA_BEANS, PLANTS * 1.4)
	make(Items.CRIMSON_FUNGUS, PLANTS * 1.161)
	make(Items.WARPED_FUNGUS, PLANTS * 1.161)
	make(Items.CRIMSON_ROOTS, GRASS * 1.161)
	make(Items.WARPED_ROOTS, GRASS * 1.161)
	make(Items.NETHER_SPROUTS, GRASS * 0.8)
	make(Items.WEEPING_VINES, GRASS * 1.161)
	make(Items.TWISTING_VINES, GRASS * 1.261)
	make(Items.DEAD_BUSH, GRASS * 0.64)

	make(Items.TALL_GRASS, GRASS * 1.7)
	make(Items.GRASS, GRASS)
	make(Items.FERN, GRASS * 1.02)
	make(Items.LARGE_FERN, GRASS * 1.91)
	make(Items.SEAGRASS, GRASS * 0.94)

	make(Items.KELP, PLANTS * 0.94)
	make(Items.MOSS_BLOCK, PLANTS * 1.4)
	make(Items.HANGING_ROOTS, PLANTS * 0.8)
	make(Items.BIG_DRIPLEAF, PLANTS * 1.15)
	make(Items.SMALL_DRIPLEAF, PLANTS * 1.1)
	make(Items.AZALEA, PLANTS * 1.3)
	make(Items.FLOWERING_AZALEA, PLANTS * 1.4)
	make(Items.FLOWERING_AZALEA_LEAVES, PLANTS * 1.5)

	make(Items.BONE, BONES)
	make(Items.STRING, STRINGS)
	make(Items.FEATHER, FEATHERS)

	make(Items.PORKCHOP, ADVANCED_ORGANICS * 0.65)
	make(Items.BEEF, ADVANCED_ORGANICS * 0.65)
	make(Items.LEATHER, ADVANCED_ORGANICS * 0.7)
	make(Items.EGG, ADVANCED_ORGANICS * 0.75)
	make(Items.HONEYCOMB, ADVANCED_ORGANICS * 1.15)
	make(Items.BLAZE_ROD, ADVANCED_ORGANICS * 2.3)
	make(Items.ENDER_PEARL, ADVANCED_ORGANICS * 6.5)

	make(Items.SLIME_BALL, SIMPLE_ORGANICS * 1.7)
	make(Items.INK_SAC, SIMPLE_ORGANICS * 1.4)
	make(Items.GLOW_INK_SAC, SIMPLE_ORGANICS * 2.1)
	make(Items.NAUTILUS_SHELL, SIMPLE_ORGANICS * 2.3)
	make(Items.HEART_OF_THE_SEA, SIMPLE_ORGANICS * 4)

	make(Items.MUSIC_DISC_13, MUSIC_DISC)
	make(Items.MUSIC_DISC_CAT, MUSIC_DISC)
	make(Items.MUSIC_DISC_BLOCKS, MUSIC_DISC)
	make(Items.MUSIC_DISC_CHIRP, MUSIC_DISC)
	make(Items.MUSIC_DISC_FAR, MUSIC_DISC)
	make(Items.MUSIC_DISC_MALL, MUSIC_DISC)
	make(Items.MUSIC_DISC_MELLOHI, MUSIC_DISC)
	make(Items.MUSIC_DISC_STAL, MUSIC_DISC)
	make(Items.MUSIC_DISC_STRAD, MUSIC_DISC)
	make(Items.MUSIC_DISC_WARD, MUSIC_DISC)
	make(Items.MUSIC_DISC_11, MUSIC_DISC)
	make(Items.MUSIC_DISC_WAIT, MUSIC_DISC)
	make(Items.MUSIC_DISC_OTHERSIDE, MUSIC_DISC)
	make(Items.MUSIC_DISC_PIGSTEP, MUSIC_DISC)

	make(Items.TUBE_CORAL, CORALS)
	make(Items.BRAIN_CORAL, CORALS)
	make(Items.BUBBLE_CORAL, CORALS)
	make(Items.FIRE_CORAL, CORALS)
	make(Items.HORN_CORAL, CORALS)

	make(Items.SEA_PICKLE, CORALS * 1.27)

	make(Items.DEAD_BRAIN_CORAL, DEAD_CORALS)
	make(Items.DEAD_BUBBLE_CORAL, DEAD_CORALS)
	make(Items.DEAD_FIRE_CORAL, DEAD_CORALS)
	make(Items.DEAD_HORN_CORAL, DEAD_CORALS)
	make(Items.DEAD_TUBE_CORAL, DEAD_CORALS)

	derive(Items.IRON_INGOT, Items.IRON_NUGGET, 1.0 / 9.0, 1.0 / 9.0)
	derive(Items.GOLD_INGOT, Items.GOLD_NUGGET, 1.0 / 9.0, 1.0 / 9.0)
}

private fun registerDerived() {
	deriveIfMissing(Items.BROWN_MUSHROOM, Items.BROWN_MUSHROOM_BLOCK)
	deriveIfMissing(Items.RED_MUSHROOM, Items.RED_MUSHROOM_BLOCK)
	deriveIfMissing(Items.COBWEB, Items.STRING)
	deriveIfMissing(Items.FARMLAND, Items.DIRT)

	derive(Items.ANVIL, Items.CHIPPED_ANVIL, 0.75, 0.85)
	derive(Items.ANVIL, Items.DAMAGED_ANVIL, 0.6, 0.78)

	derive(Items.WHITE_CONCRETE_POWDER, Items.WHITE_CONCRETE)
	derive(Items.ORANGE_CONCRETE_POWDER, Items.ORANGE_CONCRETE)
	derive(Items.MAGENTA_CONCRETE_POWDER, Items.MAGENTA_CONCRETE)
	derive(Items.LIGHT_BLUE_CONCRETE_POWDER, Items.LIGHT_BLUE_CONCRETE)
	derive(Items.YELLOW_CONCRETE_POWDER, Items.YELLOW_CONCRETE)
	derive(Items.LIME_CONCRETE_POWDER, Items.LIME_CONCRETE)
	derive(Items.PINK_CONCRETE_POWDER, Items.PINK_CONCRETE)
	derive(Items.GRAY_CONCRETE_POWDER, Items.GRAY_CONCRETE)
	derive(Items.LIGHT_GRAY_CONCRETE_POWDER, Items.LIGHT_GRAY_CONCRETE)
	derive(Items.CYAN_CONCRETE_POWDER, Items.CYAN_CONCRETE)
	derive(Items.PURPLE_CONCRETE_POWDER, Items.PURPLE_CONCRETE)
	derive(Items.BLUE_CONCRETE_POWDER, Items.BLUE_CONCRETE)
	derive(Items.BROWN_CONCRETE_POWDER, Items.BROWN_CONCRETE)
	derive(Items.GREEN_CONCRETE_POWDER, Items.GREEN_CONCRETE)
	derive(Items.RED_CONCRETE_POWDER, Items.RED_CONCRETE)
	derive(Items.BLACK_CONCRETE_POWDER, Items.BLACK_CONCRETE)
}
