package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.*
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.math.BigDecimal
import java.util.*
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayList
import kotlin.collections.HashMap
import kotlin.properties.ReadOnlyProperty
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty

interface FieldAccess<V> {
	fun read(): V
	fun write(value: V)
}

fun interface FieldGetter<V> {
	fun invoke(field: FieldAccess<V>): V
}

fun interface FieldSetter<V> {
	fun invoke(value: V, field: FieldAccess<V>, setByRemote: Boolean)
}

interface INetworkValueCodec<V> {
	fun read(stream: DataInputStream): V
	fun write(stream: DataOutputStream, value: V)
	fun copy(value: V): V
	fun compare(a: V, b: V): Boolean
}

class NetworkValueCodec<V>(
	private val reader: (stream: DataInputStream) -> V,
	private val writer: (stream: DataOutputStream, value: V) -> Unit,
	private val copier: ((value: V) -> V) = { it },
	private val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
) : INetworkValueCodec<V> {
	override fun read(stream: DataInputStream): V {
		return reader.invoke(stream)
	}

	override fun write(stream: DataOutputStream, value: V) {
		writer.invoke(stream, value)
	}

	override fun copy(value: V): V {
		return copier.invoke(value)
	}

	override fun compare(a: V, b: V): Boolean {
		return comparator.invoke(a, b)
	}
}

val BooleanValueCodec = NetworkValueCodec(DataInputStream::readBoolean, DataOutputStream::writeBoolean)
val ByteValueCodec = NetworkValueCodec(DataInputStream::readByte, { s, v -> s.writeByte(v.toInt()) })
val ShortValueCodec = NetworkValueCodec(DataInputStream::readShort, { s, v -> s.writeShort(v.toInt()) })
val IntValueCodec = NetworkValueCodec(DataInputStream::readInt, DataOutputStream::writeInt)
val LongValueCodec = NetworkValueCodec(DataInputStream::readLong, DataOutputStream::writeLong)
val FloatValueCodec = NetworkValueCodec(DataInputStream::readFloat, DataOutputStream::writeFloat)
val DoubleValueCodec = NetworkValueCodec(DataInputStream::readDouble, DataOutputStream::writeDouble)
val ItemStackValueCodec = NetworkValueCodec(DataInputStream::readItem, DataOutputStream::writeItem, ItemStack::copy, ItemStack::isSameItemSameTags)
val ImpreciseFractionValueCodec = NetworkValueCodec(DataInputStream::readImpreciseFraction, DataOutputStream::writeImpreciseFraction)
val BigDecimalValueCodec = NetworkValueCodec(DataInputStream::readBigDecimal, DataOutputStream::writeBigDecimal)
val UUIDValueCodec = NetworkValueCodec({ s -> UUID(s.readLong(), s.readLong()) }, { s, v -> s.writeLong(v.mostSignificantBits); s.writeLong(v.leastSignificantBits) })
val VarIntValueCodec = NetworkValueCodec(DataInputStream::readVarIntLE, DataOutputStream::writeVarIntLE)

class EnumValueCodec<V : Enum<V>>(private val clazz: Class<out V>) : INetworkValueCodec<V> {
	private val values = clazz.enumConstants

	override fun read(stream: DataInputStream): V {
		val id = stream.readVarIntLE()

		if (id >= values.size) {
			throw IndexOutOfBoundsException("Unable to read enum $clazz, no such enum at index $id")
		}

		return values[id]
	}

	override fun write(stream: DataOutputStream, value: V) {
		stream.writeVarIntLE(value.ordinal)
	}

	override fun copy(value: V): V {
		return value
	}

	override fun compare(a: V, b: V): Boolean {
		return a === b
	}
}

sealed interface IField<V> : ReadOnlyProperty<Any, V> {
	fun observe()
	fun markDirty()
	val value: V

	fun write(stream: DataOutputStream)
	fun read(stream: DataInputStream)

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return value
	}
}

sealed interface IMutableField<V> : IField<V>, ReadWriteProperty<Any, V> {
	override var value: V

	override fun setValue(thisRef: Any, property: KProperty<*>, value: V) {
		this.value = value
	}

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return super.getValue(thisRef, property)
	}
}

data class MapChangeset<out K, out V>(
	val action: MapAction,
	val key: K?,
	val value: V?
)

enum class MapAction {
	CLEAR, ADD, REMOVE
}

class FieldSynchronizer {
	private val fields = ArrayList<IField<*>>()
	private val observers = ArrayList<IField<*>>()
	private val dirtyFields = ArrayList<IField<*>>()

	fun byte(
		value: Byte = 0,
		getter: FieldGetter<Byte>? = null,
		setter: FieldSetter<Byte>? = null,
	): Field<Byte> {
		return Field(value, ByteValueCodec, getter, setter)
	}

	fun bool(
		value: Boolean = false,
		getter: FieldGetter<Boolean>? = null,
		setter: FieldSetter<Boolean>? = null,
	): Field<Boolean> {
		return Field(value, BooleanValueCodec, getter, setter)
	}

	fun short(
		value: Short = 0,
		getter: FieldGetter<Short>? = null,
		setter: FieldSetter<Short>? = null,
	): Field<Short> {
		return Field(value, ShortValueCodec, getter, setter)
	}

	fun long(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,
	): Field<Long> {
		return Field(value, LongValueCodec, getter, setter)
	}

	fun float(
		value: Float = 0f,
		getter: FieldGetter<Float>? = null,
		setter: FieldSetter<Float>? = null,
	): Field<Float> {
		return Field(value, FloatValueCodec, getter, setter)
	}

	fun double(
		value: Double = 0.0,
		getter: FieldGetter<Double>? = null,
		setter: FieldSetter<Double>? = null,
	): Field<Double> {
		return Field(value, DoubleValueCodec, getter, setter)
	}

	fun uuid(
		value: UUID = UUID(0L, 0L),
		getter: FieldGetter<UUID>? = null,
		setter: FieldSetter<UUID>? = null,
	): Field<UUID> {
		return Field(value, UUIDValueCodec, getter, setter)
	}

	fun int(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
	): Field<Int> {
		return Field(value, IntValueCodec, getter, setter)
	}

	fun varInt(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
	): Field<Int> {
		return Field(value, VarIntValueCodec, getter, setter)
	}

	fun fraction(
		value: ImpreciseFraction = ImpreciseFraction.ZERO,
		getter: FieldGetter<ImpreciseFraction>? = null,
		setter: FieldSetter<ImpreciseFraction>? = null,
	): Field<ImpreciseFraction> {
		return Field(value, ImpreciseFractionValueCodec, getter, setter)
	}

	fun bigDecimal(
		value: BigDecimal = BigDecimal.ZERO,
		getter: FieldGetter<BigDecimal>? = null,
		setter: FieldSetter<BigDecimal>? = null,
	): Field<BigDecimal> {
		return Field(value, BigDecimalValueCodec, getter, setter)
	}

	fun <T : Enum<T>> enum(
		type: Class<T>,
		value: T = type.enumConstants[0],
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(type), getter, setter)
	}

	fun <T : Enum<T>> enum(
		value: T,
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(value::class.java), getter, setter)
	}

	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = true)
	}

	fun item(
		delegate: KMutableProperty0<ItemStack>
	): ObservedField<ItemStack> {
		return ObservedField(delegate, ItemStackValueCodec)
	}

	fun <K, V> map(
		keyCodec: INetworkValueCodec<K>,
		valueCodec: INetworkValueCodec<V>,
		callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
		backingMap: MutableMap<K, V> = HashMap(),
		observingBackingMap: MutableMap<K, V>? = null,
	): Map<K, V> {
		return Map(
			keyCodec = keyCodec,
			valueCodec = valueCodec,
			callback = callback,
			backingMap = backingMap,
			observingBackingMap = observingBackingMap
		)
	}

	inner class Field<V>(
		private var field: V,
		private val codec: INetworkValueCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
	) : IMutableField<V> {
		private var remote: V = codec.copy(field)

		val id = fields.size + 1

		init {
			fields.add(this)

			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return field
			}

			override fun write(value: V) {
				if (!isDirty && !codec.compare(remote, value)) {
					dirtyFields.add(this@Field)
					isDirty = true
				}

				this@Field.field = value
			}
		}

		override fun observe() {
			if (!isDirty && !codec.compare(remote, field)) {
				dirtyFields.add(this)
				isDirty = true
			}
		}

		override var value: V
			get() {
				val getter = this.getter

				if (getter != null) {
					return getter.invoke(access)
				}

				return this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
					return
				}

				if (this.field == value) {
					return
				}

				if (!isDirty && !codec.compare(remote, value)) {
					dirtyFields.add(this)
					isDirty = true
				}

				this.field = value
			}

		override fun markDirty() {
			if (!isDirty) {
				dirtyFields.add(this)
				isDirty = true
			}
		}

		override fun write(stream: DataOutputStream) {
			stream.write(id)
			codec.write(stream, field)
			isDirty = false
			remote = codec.copy(field)
		}

		override fun read(stream: DataInputStream) {
			val value = codec.read(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.field = value
		}
	}

	inner class ObservedField<V> : IMutableField<V> {
		private val codec: NetworkValueCodec<V>
		private val getter: () -> V
		private val setter: (V) -> Unit

		private var remote: V

		override var value: V
			get() = getter.invoke()
			set(value) { setter.invoke(value) }

		constructor(field: KMutableProperty0<V>, dispatcher: NetworkValueCodec<V>) {
			this.codec = dispatcher
			getter = field::get
			setter = field::set
			remote = dispatcher.copy(value)
		}

		constructor(getter: () -> V, setter: (V) -> Unit, dispatcher: NetworkValueCodec<V>) {
			this.codec = dispatcher
			this.getter = getter
			this.setter = setter
			remote = dispatcher.copy(value)
		}

		private var isDirty = false

		val id = fields.size + 1

		init {
			fields.add(this)
			observers.add(this)
		}

		override fun observe() {
			if (!isDirty && !codec.compare(remote, value)) {
				dirtyFields.add(this)
				isDirty = true
			}
		}

		override fun markDirty() {
			if (!isDirty) {
				dirtyFields.add(this)
				isDirty = true
			}
		}


		override fun write(stream: DataOutputStream) {
			stream.write(id)
			val value = value
			codec.write(stream, value)
			isDirty = false
			remote = codec.copy(value)
		}

		override fun read(stream: DataInputStream) {
			this.value = codec.read(stream)
		}
	}

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(MapAction.CLEAR.ordinal + 1) }
		private val MapActionList = MapAction.values()
		private val ClearMapChangeset = MapChangeset(MapAction.CLEAR, null, null)
	}

	inner class Map<K, V>(
		private val keyCodec: INetworkValueCodec<K>,
		private val valueCodec: INetworkValueCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val observingBackingMap: MutableMap<K, V>? = null,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null
	) : IField<MutableMap<K, V>> {
		private var isDirty = false
		private var sentAllValues = false
		private var isRemote = false

		val id = fields.size + 1

		init {
			fields.add(this)

			if (observingBackingMap != null)
				observers.add(this)
		}

		private val backlog = LinkedList<(DataOutputStream) -> Unit>()

		override fun observe() {
			if (isRemote) {
				return
			}

			val observingBackingMap = observingBackingMap

			if (observingBackingMap != null) {
				for ((key, value) in backingMap) {
					val remoteValue = observingBackingMap[key] ?: throw ConcurrentModificationException("Backing map of $this was modified externally, or $value missed a modification")

					if (!valueCodec.compare(value, remoteValue)) {
						val valueCopy = valueCodec.copy(value)

						backlog.add {
							it.write(MapAction.ADD.ordinal + 1)
							keyCodec.write(it, key)
							valueCodec.write(it, valueCopy)
						}

						observingBackingMap[key] = valueCopy

						if (!isDirty) {
							dirtyFields.add(this)
							isDirty = true
						}
					}
				}
			}
		}

		override fun markDirty() {
			if (isRemote) {
				return
			}

			if (!isDirty) {
				dirtyFields.add(this)
				isDirty = true
			}

			if (!sentAllValues) {
				for ((key, value) in backingMap) {
					val valueCopy = valueCodec.copy(value)

					backlog.add {
						it.write(MapAction.ADD.ordinal + 1)
						keyCodec.write(it, key)
						valueCodec.write(it, valueCopy)
					}

					observingBackingMap?.put(key, valueCopy)
				}

				sentAllValues = true
			}
		}

		override val value: MutableMap<K, V> = object : ObservedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote) {
					return
				}

				backlog.clear()
				observingBackingMap?.clear()
				backlog.add(ClearBacklogEntry)

				if (!isDirty) {
					dirtyFields.add(this@Map)
					isDirty = true
				}
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote) {
					return
				}

				val valueCopy = valueCodec.copy(value)

				backlog.add {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				observingBackingMap?.put(key, valueCopy)

				if (!isDirty) {
					dirtyFields.add(this@Map)
					isDirty = true
				}
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote) {
					return
				}

				val keyCopy = keyCodec.copy(key)

				backlog.add {
					it.write(MapAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				observingBackingMap?.remove(key)

				if (!isDirty) {
					dirtyFields.add(this@Map)
					isDirty = true
				}
			}
		}

		override fun write(stream: DataOutputStream) {
			stream.write(id)

			sentAllValues = false
			isDirty = false

			for (entry in backlog) {
				entry.invoke(stream)
			}

			backlog.clear()

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				backlog.clear()
				observingBackingMap?.clear()
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= MapActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (MapActionList[readAction]) {
					MapAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					MapAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(MapAction.ADD, key, value))
					}

					MapAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(MapAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	fun invalidate() {
		for (field in fields) {
			field.markDirty()
		}
	}

	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		if (observers.isNotEmpty()) {
			for (field in observers) {
				field.observe()
			}
		}

		if (dirtyFields.isEmpty()) {
			return null
		}

		val stream = FastByteArrayOutputStream()
		val dataStream = DataOutputStream(stream)

		for (field in dirtyFields) {
			field.write(dataStream)
		}

		dirtyFields.clear()
		dataStream.write(0)

		return stream
	}

	fun applyNetworkPayload(stream: DataInputStream): Int {
		var fieldId = stream.read()
		var i = 0

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IndexOutOfBoundsException("Invalid field id $fieldId")
			field.read(stream)
			fieldId = stream.read()
			i++
		}

		if (stream.read() != -1) {
			throw IllegalStateException("Stream wasn't fully drain!")
		}

		return i
	}

	fun applyNetworkPayload(stream: InputStream): Int {
		return applyNetworkPayload(DataInputStream(stream))
	}
}
